//-------------------------------------------------------------------------//
//                                                                         //
//  This benchmark is a serial C version of the NPB SP code. This C        //
//  version is developed by the Center for Manycore Programming at Seoul   //
//  National University and derived from the serial Fortran versions in    //
//  "NPB3.3-SER" developed by NAS.                                         //
//                                                                         //
//  Permission to use, copy, distribute and modify this software for any   //
//  purpose with or without fee is hereby granted. This software is        //
//  provided "as is" without express or implied warranty.                  //
//                                                                         //
//  Information on NPB 3.3, including the technical report, the original   //
//  specifications, source code, results and information on how to submit  //
//  new results, is available at:                                          //
//                                                                         //
//           http://www.nas.nasa.gov/Software/NPB/                         //
//                                                                         //
//  Send comments or suggestions for this C version to cmp@aces.snu.ac.kr  //
//                                                                         //
//          Center for Manycore Programming                                //
//          School of Computer Science and Engineering                     //
//          Seoul National University                                      //
//          Seoul 151-744, Korea                                           //
//                                                                         //
//          E-mail:  cmp@aces.snu.ac.kr                                    //
//                                                                         //
//-------------------------------------------------------------------------//

//-------------------------------------------------------------------------//
// Authors: Sangmin Seo, Jungwon Kim, Jun Lee, Jeongho Nah, Gangwon Jo,    //
//          and Jaejin Lee                                                 //
//-------------------------------------------------------------------------//

//---------------------------------------------------------------------
// program SP
//---------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <fstream>

#include "dump.h"

#include "header.h"
#include "print_results.h"


/* common /global/ */
int grid_points[3], nx2, ny2, nz2;
bool timeron;

/* common /constants/ */
__PROMISE__ tx1, tx2, tx3, ty1, ty2, ty3, tz1, tz2, tz3, 
       dx1, dx2, dx3, dx4, dx5, dy1, dy2, dy3, dy4, 
       dy5, dz1, dz2, dz3, dz4, dz5, dssp, dt, 
       ce[5][13], dxmax, dymax, dzmax, xxcon1, xxcon2, 
       xxcon3, xxcon4, xxcon5, dx1tx1, dx2tx1, dx3tx1,
       dx4tx1, dx5tx1, yycon1, yycon2, yycon3, yycon4,
       yycon5, dy1ty1, dy2ty1, dy3ty1, dy4ty1, dy5ty1,
       zzcon1, zzcon2, zzcon3, zzcon4, zzcon5, dz1tz1, 
       dz2tz1, dz3tz1, dz4tz1, dz5tz1, dnxm1, dnym1, 
       dnzm1, c1c2, c1c5, c3c4, c1345, conz1, c1, c2, 
       c3, c4, c5, c4dssp, c5dssp, dtdssp, dttx1, bt,
       dttx2, dtty1, dtty2, dttz1, dttz2, c2dttx1, 
       c2dtty1, c2dttz1, comz1, comz4, comz5, comz6, 
       c3c4tx3, c3c4ty3, c3c4tz3, c2iv, con43, con16;

/* common /fields/ */
__PROMISE__ u      [KMAX][JMAXP+1][IMAXP+1][5];
__PROMISE__ us     [KMAX][JMAXP+1][IMAXP+1];
__PROMISE__ vs     [KMAX][JMAXP+1][IMAXP+1];
__PROMISE__ ws     [KMAX][JMAXP+1][IMAXP+1];
__PROMISE__ qs     [KMAX][JMAXP+1][IMAXP+1];
__PROMISE__ rho_i  [KMAX][JMAXP+1][IMAXP+1];
__PROMISE__ speed  [KMAX][JMAXP+1][IMAXP+1];
__PROMISE__ square [KMAX][JMAXP+1][IMAXP+1];
__PROMISE__ rhs    [KMAX][JMAXP+1][IMAXP+1][5];
__PROMISE__ forcing[KMAX][JMAXP+1][IMAXP+1][5];

/* common /work_1d/ */
__PROMISE__ cv  [PROBLEM_SIZE];
__PROMISE__ rhon[PROBLEM_SIZE];
__PROMISE__ rhos[PROBLEM_SIZE];
__PROMISE__ rhoq[PROBLEM_SIZE];
__PROMISE__ cuf [PROBLEM_SIZE];
__PROMISE__ q   [PROBLEM_SIZE];
__PROMISE__ ue [PROBLEM_SIZE][5];
__PROMISE__ buf[PROBLEM_SIZE][5];

/* common /work_lhs/ */
__PROMISE__ lhs [IMAXP+1][IMAXP+1][5];
__PROMISE__ lhsp[IMAXP+1][IMAXP+1][5];
__PROMISE__ lhsm[IMAXP+1][IMAXP+1][5];


int main(int argc, char *argv[])
{
  cadna_init(0);  int i, niter, step, n3;
  __PROMISE__ mflops, t, tmax, trecs[t_last+1];
  bool verified;
  char Class;
  char *t_names[t_last+1];

  //---------------------------------------------------------------------
  // Read input file (if it exists), else take
  // defaults from parameters
  //---------------------------------------------------------------------
  FILE *fp;
  if ((fp = fopen("timer.flag", "r")) != NULL) {
    timeron = true;
    t_names[t_total] = "total";
    t_names[t_rhsx] = "rhsx";
    t_names[t_rhsy] = "rhsy";
    t_names[t_rhsz] = "rhsz";
    t_names[t_rhs] = "rhs";
    t_names[t_xsolve] = "xsolve";
    t_names[t_ysolve] = "ysolve";
    t_names[t_zsolve] = "zsolve";
    t_names[t_rdis1] = "redist1";
    t_names[t_rdis2] = "redist2";
    t_names[t_tzetar] = "tzetar";
    t_names[t_ninvr] = "ninvr";
    t_names[t_pinvr] = "pinvr";
    t_names[t_txinvr] = "txinvr";
    t_names[t_add] = "add";
    fclose(fp);
  } else {
    timeron = false;
  }

  printf("\n\n NAS Parallel Benchmarks (NPB3.3-SER-C) - SP Benchmark\n\n");

  if ((fp = fopen("inputsp.data", "r")) != NULL) {
    int result;
    printf(" Reading from input file inputsp.data\n");
    result = fscanf(fp, "%d", &niter);
    while (fgetc(fp) != '\n');
    result = fscanf(fp, "%lf", &dt);
    while (fgetc(fp) != '\n');
    result = fscanf(fp, "%d%d%d", &grid_points[0], &grid_points[1], &grid_points[2]);
    fclose(fp);
  } else {
    printf(" No input file inputsp.data. Using compiled defaults\n");
    niter = NITER_DEFAULT;
    dt    = DT_DEFAULT;
    grid_points[0] = PROBLEM_SIZE;
    grid_points[1] = PROBLEM_SIZE;
    grid_points[2] = PROBLEM_SIZE;
  }

  printf(" Size: %4dx%4dx%4d\n", 
      grid_points[0], grid_points[1], grid_points[2]);
  printf(" Iterations: %4d    dt: %s\n", niter, strp(dt));
  printf("\n");

  if ((grid_points[0] > IMAX) ||
      (grid_points[1] > JMAX) ||
      (grid_points[2] > KMAX) ) {
    printf(" %d, %d, %d\n", grid_points[0], grid_points[1], grid_points[2]);
    printf(" Problem size too big for compiled array sizes\n");


    cadna_end();
    return 0;
  }
  nx2 = grid_points[0] - 2;
  ny2 = grid_points[1] - 2;
  nz2 = grid_points[2] - 2;

  set_constants();

  for (i = 1; i <= t_last; i++) {
    timer_clear(i);
  }

  exact_rhs();

  initialize();

  //---------------------------------------------------------------------
  // do one time step to touch all code, and reinitialize
  //---------------------------------------------------------------------
  adi();
  initialize();

  for (i = 1; i <= t_last; i++) {
    timer_clear(i);
  }
  timer_start(1);

  for (step = 1; step <= niter; step++) {
    if ((step % 20) == 0 || step == 1) {
      printf(" Time step %4d\n", step);
    }

    adi();
  }

  timer_stop(1);
  tmax = timer_read(1);

  verify(niter, &Class, &verified);

  if (tmax != 0.0) {
    n3 = grid_points[0]*grid_points[1]*grid_points[2];
    t = (grid_points[0]+grid_points[1]+grid_points[2])/3.0;
    mflops = (881.174 * (__PROMISE__)n3
             - 4683.91 * (t * t)
             + 11484.5 * t
             - 19272.4) * (__PROMISE__)niter / (tmax*1000000.0);
  } else {
    mflops = 0.0;
  }

  /* print_results("SP", Class, grid_points[0],  */
  /*               grid_points[1], grid_points[2], niter,  */
  /*               tmax, mflops, "          floating point",  */
  /*               verified, NPBVERSION,COMPILETIME, CS1, CS2, CS3, CS4, CS5,  */
  /*               CS6, "(none)"); */




  __PROMISE__ xi, eta, zeta, u_exact[5], add;
  for (int k = 0; k <= grid_points[2]-1; k++) {
    zeta = k * dnzm1;
    for (int j = 0; j <= grid_points[1]-1; j++) {
      eta = j * dnym1;
      for (i = 0; i <= grid_points[0]-1; i++) {
        xi = i * dnxm1;
        exact_solution(xi, eta, zeta, u_exact);
	dump(u_exact,5);

        for (int m = 0; m < 5; m++) {
	  //std::cout << u_exact[m] << std::endl;
	  //u_exact[m].dump("result.bin");
        }
      }
    }
  }
  for (int k = 1; k <= nz2; k++) {
    for (int j = 1; j <= ny2; j++) {
      for (i = 1; i <= nx2; i++) {
        for (int m = 0; m < 5; m++) {

        }
      }
    }
  }

  cadna_end();

  return 0;

  cadna_end();
}

