#!/usr/bin/python
# -*- coding: utf-8 -*-


#Copyright 2016 Romain PICOT
#
#This file is part of PROMISE.
#
#    PROMISE is free software: you can redistribute it and/or modify it
#    under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    PROMISE is distributed in the hope that it will be useful, but WITHOUT
#    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
#    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
#    Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with PROMISE. If not, see
#    <http://www.gnu.org/licenses/>.
#

import DD
import commands
import string
import sys
import os
import shutil

compil = 0
execution = 0

class Promise(DD.DD):
    def __init__(self, nab = [], index = 0, path_or = "./"):
        """
        Initialisation of PROMISE

        :param nab: array to test
        :param index: array size
        :param path: the execution path of PROMISE
        Initialize Promise class and Delta Debug algorithm
        """
        self.nabla = nab
        self.index_max = index
        self.path = path_or
        DD.DD.__init__(self)
  

    def search_index(self,listing,key):
        """
        Search if a variable is in the array to test or not

        :param listing: list of all elements
        :param key: key to be searched
        :return: index of key or None
        """
        if listing == []:
            return None
        x1 ,x2, x3 = zip(*listing)
        if key in x1:
            return x1.index(key)
        else:
            return None
              
    def test_result(self):
        """
        Test a confguration and return the result

        :return: FAIL if the configuration failed or PASS otherwise
        """
        from pprint import pprint as p
        from test_promise import compile
        from test_promise import test_significant_digit
        global compil
        global execution
        compil = compil + 1
        # Compilation
        tmp = test_promise.compile()
        if tmp == "FAIL":
            return "FAIL"

        execution = execution + 1
        # Run program
        tmp = test_promise.test_significant_digit(self.path)
        return tmp

    def test_result_high(self):
        """
        Test the highest precision confguration and return the result

        :return: FAIL if the configuration failed or PASS otherwise
        """        
        from pprint import pprint as p
        from test_promise import compile
        from test_promise import test_significant_digit_high
        global compil
        global execution
        compil = compil + 1
        #compilation
        tmp = test_promise.compile()
        if tmp == "FAIL":
            return "FAIL"

        execution = execution + 1
        #Run program
        tmp = test_promise.test_significant_digit_high(self.path)
        return tmp

    
    def create_file(self, deltas):
        """
        Create the file that will be test

        :param deltas: array of types that will be tested in float_st
        """
        input = ""
        x1 ,x2 ,x3 = zip(*deltas)
        y1 ,y2, y3 = zip(*self.nabla)
        file_memory = y3[0]
        if not os.path.exists('test/'):
            os.makedirs('test')
            shutil.copy('dump.h', 'test/')
        out = open('test/' + y3[0], 'w')
        for i in range(1, self.index_max):
            ind = self.search_index(deltas, i)
            if ind != None:
                if file_memory != x3[int(ind)]:
                    out.close()
                    out = open('test/' + x3[int(ind)],'w')
                    file_memory = x3[int(ind)]
                out.write(x2[int(ind)])
            else:
                if file_memory != y3[i-1]:
                    out.close()
                    out = open('test/' + y3[i-1],'w')
                    file_memory = y3[i-1]
                out.write(y2[i-1])
                self.search_index(self.nabla,i)
        out.close()

    def create_high(self, file_input, high):
        """
        Create the highest precision configuration file that will be test

        :param file_input: File that should be tested
        :param high: highest precision type
        """
        re_tag = re.compile('(.*?)(__PROMISE__)(.*)', re.DOTALL)
        for file in file_input:
            input = ""
            out = open('tmp', 'w')
            fread = open(file)
            for line in fread:
                mat = re_tag.match(line)
                while mat != None:
                    input = input + mat.group(1) + high
                    line = mat.group(3)
                    mat = re_tag.match(line)
                input = input + line
            out.write(input)
            out.close()
            if not os.path.exists('test/'):
                os.makedirs('test')
                shutil.copy('dump.h', 'test/')
            (status, output) =commands.getstatusoutput(path + "/cadnaizer_ref " + os.getcwd() +"/tmp -o "+ os.getcwd() +"/test/" + file)

    def create_conf(self, deltas, low):
        """
        Create the lowest precision configuration file that will be test

        :param deltas: array of types that will be tested in lowest precision
        :param low: lowest precision type
        """
        input = ""
        x1 ,x2 ,x3 = zip(*deltas)
        y1 ,y2, y3 = zip(*self.nabla)
        file_memory = y3[0]
        if not os.path.exists('test/'):
            os.makedirs('test/')
            shutil.copy('dump.h', 'test/')
        out = open('test/' + y3[0], 'w')
        for i in range(1, self.index_max):
            ind = self.search_index(deltas, i)
            if ind != None:
                if file_memory != x3[int(ind)]:
                    out.close()
                    out = open('test/' + x3[int(ind)],'w')
                    file_memory = x3[int(ind)]
                out.write(low)
            else:
                if file_memory != y3[i-1]:
                    out.close()
                    out = open('test/' + y3[i-1],'w')
                    file_memory = y3[i-1]
                out.write(y2[i-1])
                self.search_index(self.nabla,i)
        out.close()   


    def create_end_file(self, deltas):
        """
        Create files according to configuration found

        :param deltas: array of types that should be defined in lowest precision
        """
        input = ""
        float_cpt = 0
        double_cpt = 0
        x1 ,x2, x3 = zip(*deltas)
        y1 ,y2, y3 = zip(*self.nabla)
        re_ = re.compile('(.*)')
        re_sent= re.compile("""(.*?)(\/\*_PROMISE_old_|\/\*_PROMISE_\*\/)(.*?)(_PROMISE_old_\*\/|\/\*_PROMISE_\*\/)(.*)""", re.DOTALL)
        file_memory = y3[0]
        if not os.path.exists('res/'):
            os.makedirs('res/')
            shutil.copy('dump.h', 'test/')
        out = open('res/' + y3[0], 'w')
        for i in range(1, self.index_max):
            ind = self.search_index(deltas, i)
            if ind != None:
                if file_memory!= x3[int(ind)]:
                    out.close()
                    out = open('res/' + x3[int(ind)],'w')
                    file_memory=(x3[int(ind)])
                mat = re_.match(x2[int(ind)])
                if mat != None:
                    if mat.group(1) == "float":
                        float_cpt += 1
                        out.write(mat.group(1))
                    else:
                        out.write(x2[int(ind)])
                else:
                    out.write(x2[int(ind)])
            else:
                if file_memory!= y3[i-1]:
                    out.close()
                    out = open('res/' + y3[i-1],'w')
                    file_memory=(y3[i-1])
                mat = re_.match(y2[i-1])
                if mat != None:
                    if mat.group(1) == "double":
                        double_cpt += 1
                        out.write(mat.group(1))
                    else:
                        string = y2[i-1]
                        input = ""
                        while string != "":
                            mat = re_sent.match(string)
                            if mat != None:
                                input = input + mat.group(1)
                                if mat.group(2) == "/*_PROMISE_old_":
                                    input = input + mat.group(3)
                                    string = mat.group(5)
                                else:
                                    string = mat.group(5)
                            else:
                                input = input + string
                                string = ""
                                out.write(input)
                else:
                    string = y2[i-1]
                    input = ""
                    while string != "":
                        mat = re_sent.match(string)
                        if mat != None:
                            input = input + mat.group(1)
                            if mat.group(2) == "/*_PROMISE_old_":
                                input = input + mat.group(3)
                                string = mat.group(5)
                            else:
                                string = mat.group(5)
                        else:
                            input = input + string
                            string = ""
                    out.write(input)

                self.search_index(self.nabla,i)
        out.close()
        print "float amount: " + str(float_cpt)
        print "double amount: " + str(double_cpt)

    def _test(self, deltas):
        """
        create and test a configuration

        :param deltas: array to test
        """
        self.create_file(deltas)
        ret = self.test_result()
        if ret == "FAIL":
            return self.FAIL
        elif ret == "PASS":
            return self.PASS
        else:
            return self.UNRESOLVED

        #print self.coerce(deltas)

        
    
    def coerce(self, deltas):
        """Pretty-print the configuration"""
        input = ""
        for (index, line, code_sample) in deltas:
            input = input + line
        return input

def match_type(deltas, nabla, file_, index, code_sample):
    """
    locate the variable to test

    :param deltas: array that contains the type declaration
    :param nabla: array that contains evereything except deltas
    :param file: the tested file
    :param index: index at start
    :param code_sample: the code in file
    """
    re_tag = re.compile('(.*?)(__PROMISE__)(.*)', re.DOTALL)

    mat = re_tag.match(file_)
    while mat != None:
        nabla.append((index, mat.group(1), code_sample))
        index += 1
        tmp = 'float'
        deltas.append((index, tmp, code_sample))
        tmp = 'double'
        file_ = str(mat.group(3))
        nabla.append((index, tmp, code_sample))
        index += 1
        mat = re_tag.match(file_)
        
    nabla.append((index,str(file_), code_sample))
    index += 1
    return deltas, index    
    


if __name__ == '__main__':
    """main function"""
    import sys
    import re
    import inspect
    path  = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
    sys.path.insert(0, os.getcwd())
    import test_promise

    nabla = []
    deltas = []
    file_input = []

    index = 1
    test = 0

    if len(sys.argv) == 1:
        enter_input = True
    else:
        enter_input = False
        input = open(sys.argv[1])
        for line in input:
            line = line.rstrip('\n')
            file_input.append(line)

    while enter_input:
        user_input = raw_input('What is the name of the file you want to test? ')
        if user_input == "":
            enter_input = False
        else:
            file_input.append(user_input)

    for file in file_input:
        fread = open(file)
        for line in fread:
            deltas, index = match_type(deltas, nabla, line, index, file)
        fread.close()
            
    promise = Promise(nabla, index, path)
    
    print "try with the highest type"
    promise.create_high(file_input, "double")
    res = promise.test_result_high()
    if res != "PASS":
        print "No solution can be found, you should lower your expectation"
        exit()

    print "try with the double version"
    promise.create_conf(deltas, "double")
    res = promise.test_result()
    if res != "PASS":
        print "oh dear we are in trouble"
        exit()
        
    print "try with the lowest type"
    promise.create_conf(deltas, "float")
    res = promise.test_result()
    if res == "PASS":
        print "The lower type statisfy the requirement"
        promise.create_end_file(deltas)
        exit()



    print "Try a mixed solution"
    c = promise.ddmax(deltas)	# Invoke DD Algorithm
    delta = []
    for elem in deltas:
        if elem not in c:
            delta.append(elem)

    try:
        promise.create_end_file(delta)
        print "A solution have been found!"
        print "Look the file 'res' folder to see it"
        if '-outCadna' in sys.argv:
            for file in file_input:
                (status, output) =commands.getstatusoutput(path + "/cadnaizer " + 'res/' + file +" -o "+ os.getcwd() +"/res/tmp")
                (status, output) =commands.getstatusoutput("mv res/tmp " + os.getcwd() +"/res/" + file)
        else:
            for file in file_input:
                (status, output) =commands.getstatusoutput("""sed -i 's/cadna_init(0);//g' res/""" + file)
                (status, output) =commands.getstatusoutput("""sed -i 's/dump(.*);//g' res/""" + file)
    except ValueError:
        print "No solution found, you should lower your expectation"


# Local Variables:
# mode: python
# End:
